//
// Kernel Debugger Protocol Module - VMware Shared Memory
// Copyright (C) 2007 Ken Johnson (Skywing)
// http://www.nynaeve.net
// skywing@valhallalegends.com
//
// The kdvmware module is a simple driver that once loaded, performs an export
// table patch on the NTOS imports on kdcom.dll in order to gain control of the
// kernel debugger interface.  When sending or receiving a packet, we write
// bits to a shared memory region and then suspend the VM until the host-side
// debugger can process the event.  This is accomplished by repurposing the OS
// not loaded magic VMware backdoor API as a kernel debugger event notification
// mechanism.
//
// Host-side, the debugger engine has been patched to send and receive packets
// from the shared memory region.  VMware VMX is patched to gain control of the
// OS not loaded backdoor API.
//

#include <ntifs.h>
#include <Aux_klib.h>
#include <ntstrsafe.h>
#include "kdvmware.h"

#define DISABLE_ON_BUGCHECK    0

typedef KDVM_DATA_BLOCK volatile * PKDVM_DATA_BLOCK_V;

PKDVM_DATA_BLOCK_V        KdVmDataBlock;
BOOLEAN                   KdVmSuspendOnRecv = TRUE;
ULONG64                   KdVmNtosBase;
ULONG64                   KdVmNtosLimit;
ULONG64                   KdVmKdComBase;
ULONG64                   KdVmKdComLimit;
BOOLEAN                   KdVmInBugCheck;
KBUGCHECK_CALLBACK_RECORD KdVmBugCheckCallBackRecord;
extern PUCHAR            *KiBugCheckData;


#ifdef _AMD64_

const USHORT KdCompGetByteOpcodes[] =
{
	0x48, 0x83, 0xEC, 0x28,                   //  sub     rsp, 28h
	0x48, 0x8B, 0xD1,                         //  mov     rdx, rcx
	0x48, 0x8D, 0x0D, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, //  lea     rcx, Port
	0x41, 0xB0, 0x01,                         //  mov     r8b, 1
	0xE8, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF,             //  call    CpGetByte
	0x0F, 0xB7, 0xC0,                         //  movzx   eax, ax
	0x48, 0x83, 0xC4, 0x28,                   //  add     rsp, 28h
	0xC3                                      //  retn
};

const USHORT KdCompPollByteOpcodes[] =
{
	0x48, 0x83, 0xEC, 0x28,                   //  sub     rsp, 28h
	0x48, 0x8B, 0xD1,                         //  mov     rdx, rcx
	0x48, 0x8D, 0x0D, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, //  lea     rcx, Port
	0x45, 0x33, 0xC0,                         //  xor     r8d, r8d
	0xE8, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF,             //  call    CpGetByte
	0x0F, 0xB7, 0xC0,                         //  movzx   eax, ax
	0x48, 0x83, 0xC4, 0x28,                   //  add     rsp, 28h
	0xC3                                      //  retn
};

const USHORT KdCompPutByteOpcodes[] =
{
	0x48, 0x89, 0x5C, 0x24, 0x08,              //  mov     [rsp+arg_0], rbx
	0x57,                                      //  push    rdi
	0x48, 0x83, 0xEC, 0x20,                    //  sub     rsp, 20h
	0x80, 0x3D, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF,  //  cmp     cs:KdCompDbgPortsPresent, 0
	0x8A, 0xD9,                                //  mov     bl, cl
	0x74, 0x6E                                 //  jz      short loc_7FF71381BF7
};

#else

const USHORT KdCompGetByteOpcodes[] =
{
0x8B, 0xFF,          
0x55,             
0x8B, 0xEC,          
0x6A, 0x01,          
0xFF, 0x75, 0x08,       
0x68, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 
0xE8, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 
0x0F, 0xB7, 0xC0,       
0x5D,             
0xC2, 0x04, 0x00  
};

const USHORT KdCompPollByteOpcodes[] = 
{
0x8B, 0xFF,          
0x55,             
0x8B, 0xEC,          
0x6A, 0x00,          
0xFF, 0x75, 0x08,       
0x68, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 
0xE8, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 
0x0F, 0xB7, 0xC0,       
0x5D,             
0xC2, 0x04, 0x00      
};

const USHORT KdCompPutByteOpcodes[] =
{
0x8B, 0xFF,          
0x55,             
0x8B, 0xEC,          
0xFF, 0x75, 0x08,       
0x68, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 
0xE8, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 
0x5D,             
0xC2, 0x04, 0x00     
};

//#error No support for this architecture presently.

#endif

VOID
KdVmEnterBugCheck(
	__in BOOLEAN InKeBugCheck2
	)
{
	KdVmInBugCheck = TRUE;

	KdVmNotifyEvent(
		KdVmOperationBugCheck
		);

	*KdDebuggerNotPresent              = TRUE;
	SharedUserData->KdDebuggerEnabled &= ~2;

	UNREFERENCED_PARAMETER( InKeBugCheck2 );
}

KD_RECV_CODE
KdReceivePacket(
	__in ULONG PacketType,
	__inout_opt PKD_BUFFER PacketData,
	__inout_opt PKD_BUFFER SecondaryData,
	__out_opt PULONG PayloadBytes,
	__inout_opt PKD_CONTEXT KdContext
	)
{
	if (!KdVmDataBlock || KdVmInBugCheck)
	{
		if (PacketType == PACKET_TYPE_KD_BREAKIN_POLL)
			return KD_RECV_CODE_TIMEOUT;
		else
			return KD_RECV_CODE_FAILED;
	}

#if DISABLE_ON_BUGCHECK
	//
	// Check to see if we are getting called during a bug check.  If so then
	// we will disable the debugger connection so that the bug check can be
	// processed.  This is to allow for debugging of the debugger link itself
	// if we get into a bad state as we'll be able to write a bug check dump.
	//

	if (*KiBugCheckData)
	{
		KdVmEnterBugCheck(TRUE);

		if (PacketType == PACKET_TYPE_KD_BREAKIN_POLL)
			return KD_RECV_CODE_TIMEOUT;
		else
			return KD_RECV_CODE_FAILED;
	}
#endif

	if (KdVmDataBlock->Parameters.PrimaryLength +
		KdVmDataBlock->Parameters.SecondaryLength > KD_MAX_DATA_BLOCK_SIZE)
	{
		KdVmEnterBugCheck(TRUE);

		KeBugCheckEx(
			BUGCODE_KDVMWARE,
			BUGID_KDVMWARE_LARGE_RECV,
			(ULONG_PTR)PacketData,
			(ULONG_PTR)SecondaryData,
			(ULONG_PTR)KdVmDataBlock
			);
	}

	//
	// Copy the request to the data block and fire it off.
	//

	KdVmDataBlock->Parameters.PacketType      = PacketType;
	KdVmDataBlock->Parameters.PrimaryLength   = PacketData ? PacketData->MaximumLength : 0;
	KdVmDataBlock->Parameters.SecondaryLength = SecondaryData ? SecondaryData->MaximumLength : 0;

	KdVmNotifyEvent(
		KdVmOperationReceivePacket
		);

	//
	// Copy the reply data to the caller.
	//

	if (KdVmDataBlock->Parameters.RecvCode == KD_RECV_CODE_OK)
	{
		if (PacketData)
		{
			PacketData->Length = KdVmDataBlock->Parameters.PrimaryLength;

			RtlCopyMemory(
				PacketData->Data,
				(PVOID)KdVmDataBlock->Parameters.Buffer,
				KdVmDataBlock->Parameters.PrimaryLength
				);
		}

		if (SecondaryData)
		{
			SecondaryData->Length = KdVmDataBlock->Parameters.SecondaryLength;

			RtlCopyMemory(
				SecondaryData->Data,
				(PVOID)(KdVmDataBlock->Parameters.Buffer + KdVmDataBlock->Parameters.PrimaryLength),
				KdVmDataBlock->Parameters.SecondaryLength
				);
		}

		if (PayloadBytes)
			*PayloadBytes = KdVmDataBlock->Parameters.SecondaryLength;
	}

	if (KdVmDataBlock->Parameters.Flags & KDVM_FLAG_BREAK_IN_REQUESTED)
	{
		if (KdContext)
			KdContext->BreakInRequested = TRUE;
	}

	if (KdVmDataBlock->Parameters.Flags & KDVM_FLAG_DISCONNECT_DEBUGGER)
	{
		*KdDebuggerNotPresent              = TRUE;
		SharedUserData->KdDebuggerEnabled &= ~2;
	}

	if (KdVmDataBlock->Parameters.Flags & KDVM_FLAG_CONNECT_DEBUGGER)
	{
		*KdDebuggerNotPresent              = FALSE;
		SharedUserData->KdDebuggerEnabled |= 2;
	}

	return KdVmDataBlock->Parameters.RecvCode;
}

VOID
KdSendPacket(
	__in ULONG PacketType,
	__in PKD_BUFFER PacketData,
	__in_opt PKD_BUFFER SecondaryData,
	__inout PKD_CONTEXT KdContext
	)
{
	if (!KdVmDataBlock || KdVmInBugCheck)
		return;

#if DISABLE_ON_BUGCHECK
	//
	// Check to see if we are getting called during a bug check.  If so then
	// we will disable the debugger connection so that the bug check can be
	// processed.  This is to allow for debugging of the debugger link itself
	// if we get into a bad state as we'll be able to write a bug check dump.
	//

	if (*KiBugCheckData)
	{
		KdVmEnterBugCheck(TRUE);

		return;
	}
#endif

	//
	// Copy the request to the data block and fire it off.
	//


	KdVmDataBlock->Parameters.PacketType      = PacketType;
	KdVmDataBlock->Parameters.PrimaryLength   = PacketData ? PacketData->Length : 0;
	KdVmDataBlock->Parameters.SecondaryLength = SecondaryData ? SecondaryData->Length : 0;

	if (KdVmDataBlock->Parameters.PrimaryLength +
		KdVmDataBlock->Parameters.SecondaryLength > KD_MAX_DATA_BLOCK_SIZE)
	{
		KdVmEnterBugCheck(FALSE);

		KeBugCheckEx(
			BUGCODE_KDVMWARE,
			BUGID_KDVMWARE_LARGE_SEND,
			(ULONG_PTR)PacketData,
			(ULONG_PTR)SecondaryData,
			(ULONG_PTR)KdVmDataBlock
			);
	}

	if (PacketData)
	{
		RtlCopyMemory(
			(PVOID)KdVmDataBlock->Parameters.Buffer,
			PacketData->Data,
			KdVmDataBlock->Parameters.PrimaryLength
			);
	}

	if (SecondaryData)
	{
		RtlCopyMemory(
			(PVOID)(KdVmDataBlock->Parameters.Buffer + KdVmDataBlock->Parameters.PrimaryLength),
			SecondaryData->Data,
			KdVmDataBlock->Parameters.SecondaryLength
			);
	}

	KdVmNotifyEvent(
		KdVmOperationSendPacket
		);

	//
	// Copy the reply data to the caller.
	//

	if (KdVmDataBlock->Parameters.Flags & KDVM_FLAG_BREAK_IN_REQUESTED)
	{
		if (KdContext)
			KdContext->BreakInRequested = TRUE;
	}

	if (KdVmDataBlock->Parameters.Flags & KDVM_FLAG_DISCONNECT_DEBUGGER)
	{
		*KdDebuggerNotPresent              = TRUE;
		SharedUserData->KdDebuggerEnabled &= ~2;
	}

	if (KdVmDataBlock->Parameters.Flags & KDVM_FLAG_CONNECT_DEBUGGER)
	{
		*KdDebuggerNotPresent              = FALSE;
		SharedUserData->KdDebuggerEnabled |= 2;
	}
}

NTSTATUS
KdSave(
	PVOID Unused
	)
{
	UNREFERENCED_PARAMETER( Unused );

	return STATUS_SUCCESS;
}

NTSTATUS
KdRestore(
	PVOID Unused
	)
{
	UNREFERENCED_PARAMETER( Unused );

	return STATUS_SUCCESS;
}

NTSTATUS
NTAPI
KdVmFindNtos(
	VOID
	)
{
	NTSTATUS                  Status;
	ULONG                     Size;
	ULONG                     Tries;
	ULONG                     i;
	PAUX_MODULE_EXTENDED_INFO ModuleInfo;

	//
	// Locate NTOS image base using AuxKlib - the documented method!
	//

	ModuleInfo = 0;
	Status     = AuxKlibInitialize();

	if (!NT_SUCCESS(Status))
	{
		KdPrintEx((DPFLTR_IHVDRIVER_ID, 0, "KdVmFindNtos(): AuxKlibInitialize fails, %08X\n",
			Status));

		return Status;
	}

	for (Tries = 0;
	     Tries < 3;
	     Tries += 1)
	{
		Size       = 0;
		ModuleInfo = 0;

		Status = AuxKlibQueryModuleInformation(
			&Size,
			sizeof(AUX_MODULE_EXTENDED_INFO),
			0);

		if (!NT_SUCCESS(Status))
		{
			KdPrintEx((DPFLTR_IHVDRIVER_ID, 0, "KdVmFindNtos(): AuxKlibQueryModuleInformation(size) fails, %08X\n",
				Status));

			break;
		}

		ModuleInfo = (PAUX_MODULE_EXTENDED_INFO)ExAllocatePoolWithTag(
			PagedPool,
			Size,
			KDVM_TAG);

		if (!ModuleInfo)
			break;

		Status = AuxKlibQueryModuleInformation(
			&Size,
			sizeof(AUX_MODULE_EXTENDED_INFO),
			ModuleInfo);

		if (Status == STATUS_BUFFER_TOO_SMALL)
		{
			ExFreePoolWithTag(
				ModuleInfo,
				KDVM_TAG);
			ModuleInfo = 0;
			continue;
		}

		if (!NT_SUCCESS(Status))
			break;

		Status = STATUS_NOT_FOUND;

		//
		// Search the list for NTOS...
		//

		for (i = 0;
		     i < Size / sizeof(AUX_MODULE_EXTENDED_INFO);
		     i += 1)
		{
			PCHAR ModName;

			ModName = (PCHAR)(ModuleInfo[ i ].FullPathName +
				ModuleInfo[ i ].FileNameOffset);

			//
			// N.B. Assumes we weren't booted with the equivalent of KERNEL=...
			//

			if (!_stricmp(ModName, "ntoskrnl.exe") ||
				!_stricmp(ModName, "ntkrnlpa.exe") ||
				!_stricmp(ModName, "krnltest.exe"))
			{
				KdVmNtosBase  = (ULONG64)ModuleInfo[ i ].BasicInfo.ImageBase;
				KdVmNtosLimit = KdVmNtosBase + ModuleInfo[ i ].ImageSize;

				KdPrintEx((DPFLTR_IHVDRIVER_ID, 0, "KdVmFindNtos(): nt = %p %p\n",
					(PVOID)KdVmNtosBase,
					(PVOID)KdVmNtosLimit));

				if (KdVmNtosBase && KdVmKdComBase)
				{
					Status = STATUS_SUCCESS;
					break;
				}
			}
			else if (!_stricmp(ModName, "kdcom.dll"))
			{
				KdVmKdComBase  = (ULONG64)ModuleInfo[ i ].BasicInfo.ImageBase;
				KdVmKdComLimit = KdVmKdComBase + ModuleInfo[ i ].ImageSize;

				KdPrintEx((DPFLTR_IHVDRIVER_ID, 0, "KdVmFindNtos(): kdcom = %p %p (%p) (%s) (%p)\n",
					(PVOID)KdVmKdComBase,
					(PVOID)KdVmKdComLimit,
					ModuleInfo,
					ModName,
					ModName));

				KdBreakPoint();

				if (KdVmNtosBase && KdVmKdComBase)
				{
					Status = STATUS_SUCCESS;
					break;
				}
			}
		}

		if (NT_SUCCESS(Status))
			break;

		KdPrintEx((DPFLTR_IHVDRIVER_ID, 0, "KdVmFindNtos(): Couldn't find ntoskrnl in loaded module list...\n"));

		if (ModuleInfo)
		{
			ExFreePoolWithTag(
				ModuleInfo,
				KDVM_TAG);
			ModuleInfo = 0;
		}
	}

	if (ModuleInfo)
		ExFreePoolWithTag(
			ModuleInfo,
			KDVM_TAG);

	return Status;
}

NTSTATUS
KdVmInitializeHostCommunication(
	VOID
	)
{
	//
	// Establish a communications link with the host.  This involves allocating
	// and initializing our memory region and trapping to the host until we get
	// a successful response.
	//

	NTSTATUS         Status;
	PHYSICAL_ADDRESS MaxAddr;
	ULONG            Tries;

	MaxAddr.LowPart  = 0xFFFFFFFF;
	MaxAddr.HighPart = 0xFFFFFFFF;

	Tries            = 0;

	KdVmDataBlock = (PKDVM_DATA_BLOCK_V)MmAllocateContiguousMemory(
		sizeof(KDVM_DATA_BLOCK),
		MaxAddr
		);

	if (!KdVmDataBlock)
	{
		KdPrintEx((DPFLTR_IHVDRIVER_ID, 0, "KdVmInitializeHostCommunication(): Failed to allocate contiguous physical memory.\n"));

		Status = STATUS_NO_MEMORY;
		return Status;
	}

	//
	// We need to initialize the KDVM data block.  Do so now.
	//

	RtlZeroMemory(
		(PVOID)KdVmDataBlock,
		sizeof(KDVM_DATA_BLOCK)
		);

	RtlCopyMemory(
		(PVOID)KdVmDataBlock->Magic,
		KDVM_SIGNATURE_INITIALIZING,
		KDVM_SIGNATURE_LENGTH
		);

	KdVmDataBlock->Magic[ 62 ] = 0xAA;

	KdVmDataBlock->Length         = sizeof(KDVM_DATA_BLOCK);

	//
	// We'll stop here until the host acknowledges us.
	//

	while (!KdVmDataBlock->HostLinkActive && Tries < 5)
	{
		KdPrintEx((DPFLTR_IHVDRIVER_ID, 0, "KdVmInitializeHostCommunication(): Trying to connect to host (%lu)...\n", Tries));

		Tries += 1;

		KdVmNotifyEvent(KdVmOperationInitialize);
	}

	if (!KdVmDataBlock->HostLinkActive)
	{
		KdPrintEx((DPFLTR_IHVDRIVER_ID, 0, "KdVmInitializeHostCommunication(): Failed to connect to host!  Unloading driver image.\n"));

		RtlZeroMemory(
			(PVOID)KdVmDataBlock,
			sizeof(KDVM_DATA_BLOCK)
			);

		MmFreeContiguousMemory((PVOID)KdVmDataBlock);
		KdVmDataBlock = 0;

		return STATUS_PORT_DISCONNECTED;
	}

	//
	// We're all good to go.  Let's return success.
	//

	return STATUS_SUCCESS;
}

PVOID
KdVmGetProcAddress(
	__in ULONG64 ImageBase,
	__in_opt PIMAGE_EXPORT_DIRECTORY ExportDir,
	__in CONST CHAR *ExportName
	)
{
	PULONG                  Names;
	PUSHORT                 Ordinals;
	PULONG                  Functions;
	ULONG                   i;

	if (!ExportDir)
	{
		ExportDir = AuxKlibGetImageExportDirectory(
			(PVOID)ImageBase);

		if (!ExportDir)
		{
			KdPrintEx((DPFLTR_IHVDRIVER_ID, 0, "KdVmGetProcAddress(): Couldn't find export directory for image.\n"));
			return 0;
		}
	}

	Names     = (PULONG)(ImageBase + ExportDir->AddressOfNames);
	Ordinals  = (PUSHORT)(ImageBase + ExportDir->AddressOfNameOrdinals);
	Functions = (PULONG)(ImageBase + ExportDir->AddressOfFunctions);

	for (i = 0;
	     i < ExportDir->NumberOfNames;
	     i += 1)
	{
		PCHAR Name = (PCHAR)(KdVmKdComBase + Names[ i ]);

		if (!strcmp(Name, ExportName))
		{
			USHORT Ord      = Ordinals[ i ];
			PVOID  Function = (PVOID)(KdVmKdComBase + Functions[ Ord ]);

			return Function;
		}
	}

	KdPrintEx((DPFLTR_IHVDRIVER_ID, 0, "KdVmGetProcAddress(): Failed to resolve symbol '%s'.\n",
		ExportName
		));

	return 0;
}

NTSTATUS
KdVmLockVaForWrite(
	__in PVOID Va,
	__in ULONG Length,
	__out PREPROTECT_CONTEXT ReprotectContext
	)
{
	NTSTATUS Status;

	Status = STATUS_SUCCESS;

	ReprotectContext->Mdl      = 0;
	ReprotectContext->LockedVa = 0;

	ReprotectContext->Mdl = IoAllocateMdl(
		Va,
		Length,
		FALSE,
		FALSE,
		0
		);

	if (!ReprotectContext->Mdl)
	{
		KdPrintEx((DPFLTR_IHVDRIVER_ID, 0, "KdVmLockVaForWrite(): IoAllocateMdl fails.\n"));
		return STATUS_INSUFFICIENT_RESOURCES;
	}

	//
	// Retrieve a locked VA mapping.
	//

	__try
	{
		MmProbeAndLockPages(
			ReprotectContext->Mdl,
			KernelMode,
			IoModifyAccess
			);
	}
	__except (EXCEPTION_EXECUTE_HANDLER)
	{
		KdPrintEx((DPFLTR_IHVDRIVER_ID, 0, "KdVmLockVaForWrite(): MmProbeAndLockPages fails (%08X).\n", GetExceptionCode()));
	}

	ReprotectContext->LockedVa = (PUCHAR)MmMapLockedPagesSpecifyCache(
		ReprotectContext->Mdl,
		KernelMode,
		MmCached,
		0,
		FALSE,
		NormalPagePriority
		);

	if (!ReprotectContext->LockedVa)
	{
		KdPrintEx((DPFLTR_IHVDRIVER_ID, 0, "KdVmLockVaForWrite(): MmMapLockedPagesSpecifyCache fails.\n"));

		IoFreeMdl(
			ReprotectContext->Mdl
			);

		ReprotectContext->Mdl = 0;

		return STATUS_ACCESS_VIOLATION;
	}

	//
	// Reprotect.
	//

	Status = MmProtectMdlSystemAddress(
		ReprotectContext->Mdl,
		PAGE_EXECUTE_READWRITE
		);

	if (!NT_SUCCESS(Status))
	{
		KdPrintEx((DPFLTR_IHVDRIVER_ID, 0, "KdVmLockVaForWrite(): MmProtectMdlSystemAddress fails (%08X).\n",
			Status));

		MmUnmapLockedPages(
			ReprotectContext->LockedVa,
			ReprotectContext->Mdl
			);
		MmUnlockPages(
			ReprotectContext->Mdl
			);
		IoFreeMdl(
			ReprotectContext->Mdl
			);

		ReprotectContext->LockedVa = 0;
		ReprotectContext->Mdl      = 0;
	}

	return Status;
}

NTSTATUS
KdVmUnlockVaForWrite(
	__in PREPROTECT_CONTEXT ReprotectContext
	)
{
	if (ReprotectContext->LockedVa)
	{
		MmUnmapLockedPages(
			ReprotectContext->LockedVa,
			ReprotectContext->Mdl
			);
		MmUnlockPages(
			ReprotectContext->Mdl
			);
		IoFreeMdl(
			ReprotectContext->Mdl
			);

		ReprotectContext->LockedVa = 0;
		ReprotectContext->Mdl      = 0;
	}

	return STATUS_SUCCESS;
}

NTSTATUS
KdVmPatchSymbolToJump(
	__in PVOID Symbol,
	__in PVOID Target
	)
{
	REPROTECT_CONTEXT Reprot;
	NTSTATUS          Status;

	Status = KdVmLockVaForWrite(
		Symbol,
		15,
		&Reprot
		);

	if (!NT_SUCCESS(Status))
		return Status;

#ifdef _AMD64_

	Symbol = (PVOID)Reprot.LockedVa;

	*(((PUCHAR)Symbol) + 0)             = 0x48;
	*(((PUCHAR)Symbol) + 1)             = 0xB8;
	*(PULONGLONG)(((PUCHAR)Symbol) + 2) = (ULONGLONG)Target;
	*(((PUCHAR)Symbol) + 10)            = 0xFF;
	*(((PUCHAR)Symbol) + 11)            = 0xE0;

#elif defined(_X86_)

	*(((PUCHAR)Reprot.LockedVa) + 0)             = 0xE9;
	*(PULONG_PTR)(((PUCHAR)Reprot.LockedVa) + 1) = (ULONG_PTR)Target - (ULONG_PTR) (((PUCHAR)Symbol) + 5 );

#else

#error No support for that architecture yet.

#endif

	KdVmUnlockVaForWrite(
		&Reprot
		);

	return STATUS_SUCCESS;
}

VOID
KdVmNotifyEvent(
	KDVM_OPERATION Operation
	)
{
	//
	// Tell the host what operation we're performing for debugging and control
	// purposes.  This allows the host to know that we're still alive as we
	// unset the operation code after every transaction.
	//
	// We also unset the HostLinkActive flag for the duration of the
	// transaction so that we can know that the host side code processed our
	// request properly.  If they did not, then that's bad and we'll want to
	// reconnect the link.
	//

	KdVmDataBlock->Operation      = Operation;
	KdVmDataBlock->HostLinkActive = FALSE;

	KdVmpNotifyEvent();

//	while (KdVmDataBlock->Operation != KdVmOperationNone)
//		KeStallExecutionProcessor(100);

	//
	// If the host isn't linked anymore and we weren't already trying to do an
	// initialization operation, then try to reconnect once right now.
	//
	// N.B. There cannot be any debug prints here as any debug prints would
	// not succeed as the kernel debugger connection is down at this point.
	//

	if (!KdVmDataBlock->HostLinkActive && 
		Operation != KdVmOperationInitialize)
	{
		RtlCopyMemory(
			(PVOID)KdVmDataBlock->Magic,
			KDVM_SIGNATURE_INITIALIZING,
			KDVM_SIGNATURE_LENGTH
			);

		KdVmDataBlock->Magic[ 62 ] = 0xAA;

//		RtlStringCbCopyA(
//			(PCHAR)KdVmDataBlock->DebugString,
//			128,
//			"VM Host Disconnect!"
//			);

		KdVmNotifyEvent(
			KdVmOperationInitialize
			);
	}

	//
	// If we're requested to initiate a bug check, and the system is not
	// already processing a bug check, then initiate one now.
	//

	if (KdVmDataBlock->BugCheckRequested &&
		!KdVmInBugCheck                  &&
		Operation != KdVmOperationBugCheck)
	{
		KdVmEnterBugCheck(FALSE);

		KeBugCheckEx(
			BUGCODE_KDVMWARE,
			BUGID_KDVMWARE_USER_REQUEST,
			(ULONG_PTR)*KiBugCheckData,
			(ULONG_PTR)KdVmNtosBase,
			(ULONG_PTR)KdVmDataBlock
			);
	}

	KdVmDataBlock->Operation = KdVmOperationNone;

}

NTSTATUS
KdVmPatchKdCom(
	VOID
	)
{
	PUCHAR                  p;
	ULONG                   Patches;
	PVOID                   KdSendPacket_;
	PVOID                   KdReceivePacket_;
	PVOID                   KdSave_;
	PVOID                   KdRestore_;

	if (!KdVmKdComBase)
		return STATUS_PROCEDURE_NOT_FOUND;

	KdSendPacket_    = KdVmGetProcAddress(
		KdVmKdComBase,
		0,
		"KdSendPacket"
		);
	KdReceivePacket_ = KdVmGetProcAddress(
		KdVmKdComBase,
		0,
		"KdReceivePacket"
		);
	KdSave_          = KdVmGetProcAddress(
		KdVmKdComBase,
		0,
		"KdSave"
		);
	KdRestore_       = KdVmGetProcAddress(
		KdVmKdComBase,
		0,
		"KdRestore"
		);

	if (!KdSendPacket_    ||
		!KdReceivePacket_ ||
		!KdSave_          ||
		!KdRestore_       
		)
	{
		KdPrintEx((DPFLTR_IHVDRIVER_ID, 0, "KdVmPatchKdCom(): Missing export (%p, %p, %p, %p).\n",
			KdSendPacket_,
			KdReceivePacket_,
			KdSave_,
			KdRestore_
			));

		return STATUS_PROCEDURE_NOT_FOUND;
	}

	//
	// Patch the KdCom exports that we are caring about.
	//

	KdBreakPoint();

	KdVmPatchSymbolToJump(
		KdSendPacket_,
		KdSendPacket
		);

	KdVmPatchSymbolToJump(
		KdReceivePacket_,
		KdReceivePacket
		);

	KdVmPatchSymbolToJump(
		KdSave_,
		KdSave
		);

	KdVmPatchSymbolToJump(
		KdRestore_,
		KdRestore
		);

	return STATUS_SUCCESS;
}

/*
VOID
KdVmBugCheckCallback(
	__in PVOID Buffer,
	__in ULONG Length
	)
{
	UNREFERENCED_PARAMETER( Buffer );
	UNREFERENCED_PARAMETER( Length );

	KdVmInBugCheck = TRUE;

	KdVmNotifyEvent(
		KdVmOperationBugCheck
		);
}
*/

NTSTATUS
NTAPI
DriverEntry(
    __in PDRIVER_OBJECT  DriverObject,
    __in PUNICODE_STRING RegistryPath
    )
{
	NTSTATUS Status;

	//
	// First, find NTOS.  We'll bail if that doesn't work out (bad news)...
	//

	Status = KdVmFindNtos();

	if (!NT_SUCCESS(Status))
		return Status;

	//
	// Okay, we've got NTOS located.  Now we'll need to allocate sufficient
	// contiguous physical memory for our shared memory region.  We'll also
	// need to be initializing the shared memory region with the magic pattern
	// so that it may be located by the VMware VMX hook code.
	//

	Status = KdVmInitializeHostCommunication();

	if (!NT_SUCCESS(Status))
		return Status;

//	if (!KeRegisterBugCheckCallback(
//		&KdVmBugCheckCallbackRecord,
//		KdVmBugCheckCallback,
//		&KdVmInBugCheck,
//		sizeof( BOOLEAN ),
//		(PUCHAR)"kdvmware"))
//	{
//		KdPrintEx((DPFLTR_IHVDRIVER_ID, 0, "DriverEntry(): Failed to register bug check callback.\n"));
//	}

	//
	// Now we'll need to patch KDCOM imports in NTOS such that the image
	// exports for KDCOM point to us where relevant.
	//

	Status = KdVmPatchKdCom();

	if (!NT_SUCCESS(Status))
	{
		if (KdVmDataBlock)
		{
			RtlZeroMemory(
				(PVOID)KdVmDataBlock,
				sizeof(KDVM_DATA_BLOCK)
				);

			MmFreeContiguousMemory((PVOID)KdVmDataBlock);
			KdVmDataBlock = 0;
		}

		return Status;
	}

	return STATUS_SUCCESS;
}

