#ifdef _MSC_VER
#pragma once
#endif

#ifndef _VMKD_KDMODULES_KDVMWARE_H
#define _VMKD_KDMODULES_KDVMWARE_H

#define KDVM_TAG 'MVDK'

typedef struct _REPROTECT_CONTEXT
{
	PMDL   Mdl;
	PUCHAR LockedVa;
} REPROTECT_CONTEXT, * PREPROTECT_CONTEXT;

typedef struct _KD_BUFFER
{
	USHORT  Length;
	USHORT  MaximumLength;
	PUCHAR  Data;
} KD_BUFFER, * PKD_BUFFER;

typedef struct _KD_CONTEXT
{
	ULONG   RetryCount;
	BOOLEAN BreakInRequested;
} KD_CONTEXT, * PKD_CONTEXT;

typedef enum _KD_RECV_CODE
{
	KD_RECV_CODE_OK       = 0,
	KD_RECV_CODE_TIMEOUT  = 1,
	KD_RECV_CODE_FAILED   = 2
} KD_RECV_CODE, * PKD_RECV_CODE;

//
// \xAA, but we don't want false positives in the driver image for the string.
//
#define KDVM_SIGNATURE_INITIALIZING "KdVm Communications Back Channel Interface - (C) 2007 Skywing\x00\xBB\00"
#define KDVM_SIGNATURE_ESTABLISHED  "KdVm Communications Back Channel Interface - (C) 2007 Skywing\x00\x55\00"
#define KDVM_SIGNATURE_LENGTH 64

#define KD_MAX_DATA_BLOCK_SIZE 4000

#define KDVM_FLAG_BREAK_IN_REQUESTED  0x00000001
#define KDVM_FLAG_DISCONNECT_DEBUGGER 0x00000002
#define KDVM_FLAG_CONNECT_DEBUGGER    0x00000004

#define BUGCODE_KDVMWARE               0x0401BABE
#define BUGID_KDVMWARE_LARGE_RECV      0x00000000
#define BUGID_KDVMWARE_LARGE_SEND      0x00000001
#define BUGID_KDVMWARE_USER_REQUEST    0x00000002

typedef struct _KDVM_BUFFER_DESCRIPTOR
{
	ULONG        PacketType;
	USHORT       PrimaryLength;
	USHORT       SecondaryLength;
	ULONG        Flags;
	KD_RECV_CODE RecvCode;
	UCHAR        Buffer[ KD_MAX_DATA_BLOCK_SIZE ];
} KDVM_BUFFER_DESCRIPTOR, * PKDVM_BUFFER_DESCRIPTOR;

typedef enum _KDVM_OPERATION
{
	KdVmOperationNone,
	KdVmOperationInitialize,
	KdVmOperationReceivePacket,
	KdVmOperationSendPacket,
	KdVmOperationBugCheck,

	LastKdVmOperation
} KDVM_OPERATION, * PKDVM_OPERATION;

typedef struct _KDVM_DATA_BLOCK
{
	//
	// Magic prefix for searching for virtual VM physical page.
	//

	UCHAR                  Magic[ KDVM_SIGNATURE_LENGTH ];

	//
	// Structure length.
	//

	ULONG                  Length;

	//
	// Why we're requesting attention from the host.
	//

	KDVM_OPERATION         Operation;

	//
	// Host connection activity indicator.
	//

	BOOLEAN                HostLinkActive;

	//
	// Bug check request from host to guest code in case we get into a bad
	// state and need to get a dump to figure out what's going on.
	//

	BOOLEAN                BugCheckRequested;

	UCHAR                  Spare[ 2 ];

	//
	// Operational parameters.
	//

	KDVM_BUFFER_DESCRIPTOR Parameters;
} KDVM_DATA_BLOCK, * PKDVM_DATA_BLOCK;

C_ASSERT( sizeof(KDVM_DATA_BLOCK) <= PAGE_SIZE );

//
// Trap to the host side code that we have an event waiting.  Suspends
// VM execution until the host side code handles the event.
//

VOID
KdVmpNotifyEvent(
	VOID
	);


//
// 0 - data
// 1 - timeout
// 2 - error?
//

typedef
ULONG
(NTAPI * KdCompGetByteProc)(
	__out PUCHAR Byte
	);

typedef
ULONG
(NTAPI * KdCompPollByteProc)(
	__out PUCHAR Byte
	);

typedef
VOID
(NTAPI * KdCompPutByteProc)(
	__in UCHAR Byte
	);

DRIVER_INITIALIZE DriverEntry;

ULONG
NTAPI
KdCompGetByte(
	__out PUCHAR Byte
	);

ULONG
NTAPI
KdCompPollByte(
	__out PUCHAR Byte
	);

VOID
NTAPI
KdCompPutByte(
	__in UCHAR Byte
	);

VOID
KdVmNotifyEvent(
	KDVM_OPERATION Operation
	);

NTSTATUS
KdVmPatchKdCom(
	VOID
	);

NTSTATUS
KdVmPatchSymbolToReturn(
	__in PVOID Symbol,
	__in ULONG StackArgumentBytes
	);

NTSTATUS
KdVmPatchSymbolToJump(
	__in PVOID Symbol,
	__in PVOID Target
	);

BOOLEAN
KdVmMatchSearchPattern(
	__in CONST UCHAR* Search,
	__in CONST USHORT* Pattern,
	__in SIZE_T Length
	);

#define PACKET_TYPE_KD_BREAKIN_POLL			8

#endif
