#ifdef _MSC_VER
#pragma once
#endif

#ifndef _VMKD_KDMODULES_KDVMWARE_H
#define _VMKD_KDMODULES_KDVMWARE_H

#define KDVM_TAG 'MVDK'

typedef struct _REPROTECT_CONTEXT
{
	PMDL   Mdl;
	PUCHAR LockedVa;
} REPROTECT_CONTEXT, * PREPROTECT_CONTEXT;

typedef struct _KD_BUFFER
{
	USHORT  Length;
	USHORT  MaximumLength;
	PUCHAR  Data;
} KD_BUFFER, * PKD_BUFFER;

typedef struct _KD_CONTEXT
{
	ULONG   RetryCount;
	BOOLEAN BreakInRequested;
} KD_CONTEXT, * PKD_CONTEXT;

typedef enum _KD_RECV_CODE
{
	KD_RECV_CODE_OK       = 0,
	KD_RECV_CODE_TIMEOUT  = 1,
	KD_RECV_CODE_FAILED   = 2
} KD_RECV_CODE, * PKD_RECV_CODE;

/*
typedef struct _LOADER_PARAMETER_BLOCK
{
	LIST_ENTRY                          LoadOrderListHead;
	LIST_ENTRY                          MemoryDescriptorListHead;
	LIST_ENTRY                          BootDriverListHead;
	PVOID                               KernelStack;
	struct _KPRCB                      *Prcb;
	struct _EPROCESS                   *Process;
	struct _ETHREAD                    *Thread;
	ULONG                               RegistryLength;
	PVOID                               RegistryBase;
	PCONFIGURATION_COMPONENT_DATA       ConfigurationRoot;
	PCHAR                               ArcBootDeviceName;
	PCHAR                               ArcHalDeviceName;
	PCHAR                               NtBootPathName;
	PCHAR                               NtHalPathName;
	PCHAR                               LoadOptions;
	PNLS_DATA_BLOCK                     NlsData;
	PARC_DISK_INFORMATION               ArcDiskInformation;
	PVOID                               OemFontFile;
	PSETUP_LOADER_BLOCK                 SetupLoaderBlock;
	PLOADER_PARAMETER_EXTENSION         LoaderParameterExtension;
	union
	{
		I386_LOADER_BLOCK               I386;
		ALPHA_LOADER_BLOCK              Alpha;
		IA64_LOADER_BLOCK               Ia64;
	}                                   u;
	FIRMWARE_INFORMATION_LOADER_BLOCK   FirmwareInformation;
} LOADER_PARAMETER_BLOCK, * PLOADER_PARAMETER_BLOCK;

      +0x000 I386             : _I386_LOADER_BLOCK
         +0x000 CommonDataArea   : Ptr32 
         +0x004 MachineType      : Uint4B
         +0x008 VirtualBias      : Uint4B
      +0x000 Alpha            : _ALPHA_LOADER_BLOCK
         +0x000 PlaceHolder      : Uint4B
      +0x000 Ia64             : _IA64_LOADER_BLOCK
         +0x000 PlaceHolder      : Uint4B
   +0x068 FirmwareInformation : _FIRMWARE_INFORMATION_LOADER_BLOCK
      +0x000 FirmwareTypeEfi  : Pos 0, 1 Bit
      +0x000 Reserved         : Pos 1, 31 Bits
      +0x004 u                : <unnamed-tag>
         +0x000 EfiInformation   : _EFI_FIRMWARE_INFORMATION
            +0x000 FirmwareVersion  : Uint4B
            +0x004 VirtualEfiRuntimeServices : Ptr32 
            +0x008 SetVirtualAddressMapStatus : Int4B
            +0x00c MissedMappingsCount : Uint4B
         +0x000 PcatInformation  : _PCAT_FIRMWARE_INFORMATION
            +0x000 PlaceHolder      : Uint4B

} LOADER_PARAMETER_BLOCK, *PLOADER_PARAMETER_BLOCK;

*/

struct _LOADER_PARAMETER_BLOCK;

typedef struct _LOADER_PARAMETER_BLOCK *PLOADER_PARAMETER_BLOCK;

//
// \xAA, but we don't want false positives in the driver image for the string.
//
#define KDVM_SIGNATURE_INITIALIZING "KdVm Communications Back Channel Interface - (C) 2007 Skywing\x00\xBB\00"
#define KDVM_SIGNATURE_ESTABLISHED  "KdVm Communications Back Channel Interface - (C) 2007 Skywing\x00\x55\00"
#define KDVM_SIGNATURE_LENGTH 64

#define KD_MAX_DATA_BLOCK_SIZE 4000

#define KDVM_FLAG_BREAK_IN_REQUESTED  0x00000001
#define KDVM_FLAG_DISCONNECT_DEBUGGER 0x00000002
#define KDVM_FLAG_CONNECT_DEBUGGER    0x00000004

#define BUGCODE_KDVMWARE               0x0401BABE
#define BUGID_KDVMWARE_LARGE_RECV      0x00000000
#define BUGID_KDVMWARE_LARGE_SEND      0x00000001
#define BUGID_KDVMWARE_USER_REQUEST    0x00000002

typedef struct _KDVM_BUFFER_DESCRIPTOR
{
	ULONG        PacketType;
	USHORT       PrimaryLength;
	USHORT       SecondaryLength;
	ULONG        Flags;
	KD_RECV_CODE RecvCode;
	UCHAR        Buffer[ KD_MAX_DATA_BLOCK_SIZE ];
} KDVM_BUFFER_DESCRIPTOR, * PKDVM_BUFFER_DESCRIPTOR;

typedef enum _KDVM_OPERATION
{
	KdVmOperationNone,
	KdVmOperationInitialize,
	KdVmOperationReceivePacket,
	KdVmOperationSendPacket,
	KdVmOperationBugCheck,

	LastKdVmOperation
} KDVM_OPERATION, * PKDVM_OPERATION;

typedef struct _KDVM_DATA_BLOCK
{
	//
	// Magic prefix for searching for virtual VM physical page.
	//

	UCHAR                  Magic[ KDVM_SIGNATURE_LENGTH ];

	//
	// Structure length.
	//

	ULONG                  Length;

	//
	// Why we're requesting attention from the host.
	//

	KDVM_OPERATION         Operation;

	//
	// Host connection activity indicator.
	//

	BOOLEAN                HostLinkActive;

	//
	// Bug check request from host to guest code in case we get into a bad
	// state and need to get a dump to figure out what's going on.
	//

	BOOLEAN                BugCheckRequested;

	UCHAR                  Spare[ 2 ];

	//
	// Operational parameters.
	//

	KDVM_BUFFER_DESCRIPTOR Parameters;
} KDVM_DATA_BLOCK, * PKDVM_DATA_BLOCK;

C_ASSERT( sizeof(KDVM_DATA_BLOCK) <= PAGE_SIZE );

//
// Private support routine for notifying the VMM of an event.
//

VOID
KdVmpNotifyEvent(
	VOID
	);

//
// Trap to the host side code that we have an event waiting.  Suspends
// VM execution until the host side code handles the event.
//

VOID
KdVmNotifyEvent(
	KDVM_OPERATION Operation
	);

//
// Initialize the link with the host.
//

NTSTATUS
KdVmInitializeHostCommunication(
	VOID
	);

#define PACKET_TYPE_KD_BREAKIN_POLL			8

#endif
