#include <string.h>
#include <stdlib.h>
#include <windows.h>
#include <strsafe.h>
#include <process.h>
#include <stdarg.h>
#include "vmxpatch.h"

#define WRITE_TIMEOUT 20000
#define READ_TIMEOUT  20000

typedef struct _KD_PACKET {
	DWORD PacketLeader;
	WORD PacketType;
	WORD ByteCount;
	DWORD PacketId;
	DWORD Checksum;
} KD_PACKET, *PKD_PACKET;

//
// KD PROTOCOL CONSTANTS
//

#define PACKET_MAX_SIZE						4000
#define INITIAL_PACKET_ID					0x80800000
#define SYNC_PACKET_ID						0x00000800 // or INITIAL_PACKET_ID

//
// BREAKIN
//

#define BREAKIN_PACKET						0x62626262
#define BREAKIN_PACKET_BYTE					0x62

//
// LEADER
//

#define PACKET_LEADER						0x30303030
#define PACKET_LEADER_BYTE					0x30

#define CONTROL_PACKET_LEADER				0x69696969
#define CONTROL_PACKET_LEADER_BYTE			0x69

//
// TRAILING BYTE
//

#define PACKET_TRAILING_BYTE				0xAA

//
// PACKET TYPES
//

#define PACKET_TYPE_UNUSED					0
#define PACKET_TYPE_KD_STATE_CHANGE			1
#define PACKET_TYPE_KD_STATE_MANIPULATE		2
#define PACKET_TYPE_KD_DEBUG_IO				3
#define PACKET_TYPE_KD_ACKNOWLEDGE			4	// Control packet
#define PACKET_TYPE_KD_RESEND				5	// Control packet
#define PACKET_TYPE_KD_RESET				6	// Control packet
#define PACKET_TYPE_KD_STATE_CHANGE64		7
#define PACKET_TYPE_KD_BREAKIN_POLL			8
// XXX: Missing lots of post-win2k things besides FILE_IO here...
#define PACKET_TYPE_KD_FILE_IO              11

//
// KD API NUMBERS
//

//
// PACKET_TYPE_KD_STATE_CHANGE
//

#define DbgKdExceptionStateChange			0x00003030L
#define DbgKdLoadSymbolsStateChange			0x00003031L

//
// PACKET_TYPE_KD_STATE_MANIPULATE
//

#define DbgKdReadVirtualMemoryApi			0x00003130L
#define DbgKdWriteVirtualMemoryApi			0x00003131L
#define DbgKdGetContextApi					0x00003132L
#define DbgKdSetContextApi					0x00003133L
#define DbgKdWriteBreakPointApi				0x00003134L
#define DbgKdRestoreBreakPointApi			0x00003135L
#define DbgKdContinueApi					0x00003136L
#define DbgKdReadControlSpaceApi			0x00003137L
#define DbgKdWriteControlSpaceApi			0x00003138L
#define DbgKdReadIoSpaceApi					0x00003139L
#define DbgKdWriteIoSpaceApi				0x0000313AL
#define DbgKdRebootApi						0x0000313BL
#define DbgKdContinueApi2					0x0000313CL
#define DbgKdReadPhysicalMemoryApi			0x0000313DL
#define DbgKdWritePhysicalMemoryApi			0x0000313EL
#define DbgKdQuerySpecialCallsApi			0x0000313FL
#define DbgKdSetSpecialCallApi				0x00003140L
#define DbgKdClearSpecialCallsApi			0x00003141L
#define DbgKdSetInternalBreakPointApi		0x00003142L
#define DbgKdGetInternalBreakPointApi		0x00003143L
#define DbgKdReadIoSpaceExtendedApi			0x00003144L
#define DbgKdWriteIoSpaceExtendedApi		0x00003145L
#define DbgKdGetVersionApi					0x00003146L
#define DbgKdWriteBreakPointExApi			0x00003147L
#define DbgKdRestoreBreakPointExApi			0x00003148L
#define DbgKdCauseBugCheckApi				0x00003149L
#define DbgKdSwitchProcessor				0x00003150L
#define DbgKdPageInApi						0x00003151L
#define DbgKdReadMachineSpecificRegister	0x00003152L
#define DbgKdWriteMachineSpecificRegister	0x00003153L
#define DbgKdReadVirtualMemory64Api			0x00003154L
#define DbgKdWriteVirtualMemory64Api		0x00003155L

//
// PACKET_TYPE_KD_DEBUG_IO
//

#define DbgKdPrintStringApi					0x00003230L
#define DbgKdGetStringApi					0x00003231L






ULONG KdCompRetryCount             = 5;
ULONG KdCompNumberRetries          = 5; /* 5 */
ULONG KdCompNextPacketIdToSend     = INITIAL_PACKET_ID | SYNC_PACKET_ID;
ULONG KdCompPacketIdExpected       = INITIAL_PACKET_ID;

VOID
KdpResetState(
	VOID
	)
{
	VmxPatchDebugPrint(
		"KdpResetState(): Resetting state to initial boot values.\n");

	KdCompRetryCount         = 5;
	KdCompNumberRetries      = 5;
	KdCompNextPacketIdToSend = INITIAL_PACKET_ID | SYNC_PACKET_ID;
	KdCompPacketIdExpected   = INITIAL_PACKET_ID;
}

BOOLEAN
KdpConnectPipe(
	__in ULONG Timeout
	)
{
	OVERLAPPED Overlapped;
	ULONG      Transferred;

	if (g_PipeConnected)
		return TRUE;

	ZeroMemory(
		&Overlapped,
		sizeof( OVERLAPPED )
		);

	Overlapped.hEvent = g_VmxKdPipeEvent;

	ResetEvent(
		g_VmxKdPipeEvent
		);

	if (ConnectNamedPipe(
		g_VmxKdPipe,
		&Overlapped))
	{
		g_PipeConnected = TRUE;

		VmxPatchDebugPrint(
			"KdpConnectPipe(): Established transport connection with KD.\n");

		return TRUE;
	}

	switch (GetLastError())
	{

	case ERROR_PIPE_CONNECTED:
		g_PipeConnected = TRUE;

		VmxPatchDebugPrint(
			"KdpConnectPipe(): Picked up existing transport connection with KD.\n");

		return TRUE;

	case ERROR_IO_PENDING:
		{
			if (WaitForSingleObject(
				g_VmxKdPipeEvent,
				Timeout) != WAIT_OBJECT_0)
			{
				VmxPatchDebugPrint(
					"KdpConnectPipe(): Timed out waiting for KD connection.\n");

				CancelIo(
					g_VmxKdPipe
					);
			}

			if (GetOverlappedResult(
				g_VmxKdPipe,
				&Overlapped,
				&Transferred,
				TRUE))
			{
				g_PipeConnected = TRUE;

				VmxPatchDebugPrint(
					"KdpConnectPipe(): Established transport connection with KD (pending).\n");

				return TRUE;
			}
			else
			{
				if (GetLastError() == ERROR_PIPE_CONNECTED)
				{
					g_PipeConnected = TRUE;

					VmxPatchDebugPrint(
						"KdpConnectPipe(): Picked up existing transport connection with KD (pending).\n");

					return TRUE;
				}
				else
				{
					VmxPatchDebugPrint(
						"KdpConnectPipe(): ConnectNamedPipe fails, %lu (pending).\n",
						GetLastError());

					return FALSE;
				}
			}
		}
		break;

	default:
		VmxPatchDebugPrint(
			"KdpConnectPipe(): ConnectNamedPipe fails, %lu.\n",
			GetLastError());

		return FALSE;

	}

}

VOID
KdpDisconnectPipe(
	VOID
	)
{
	if (!g_PipeConnected)
		return;

	DisconnectNamedPipe(
		g_VmxKdPipe
		);

	g_PipeConnected = FALSE;
}

BOOLEAN
KdpWritePipe(
	__in CONST UCHAR* Buf,
	__in ULONG Length
	)
{
	OVERLAPPED Overlapped;
	ULONG      Written;

	if (!g_PipeConnected &&
		!KdpConnectPipe(WRITE_TIMEOUT))
	{
		VmxPatchDebugPrint(
			"KdpWritePipe(): Pipe is not connected\n"
			);

		return FALSE;
	}

	if (g_LogPackets)
	{
		VmxPatchDebugPrint(
			"KdpWritePipe(): Write %lu bytes to KD:\n",
			Length
			);

		dumphex((char*)Buf, Length, 4);
	}

	ZeroMemory(
		&Overlapped,
		sizeof( OVERLAPPED )
		);

	Overlapped.hEvent = g_VmxKdPipeEvent;

	ResetEvent(
		g_VmxKdPipeEvent
		);

	Written = 0;

	if (WriteFile(
		g_VmxKdPipe,
		Buf,
		Length,
		&Written,
		&Overlapped))
	{
		if (Written < Length)
		{
			VmxPatchDebugPrint(
				"KdpWritePipe(): WARNING: Short write %lu (%lu expected)\n",
				Written,
				Length
				);

			return FALSE;
		}

		return TRUE;
	}

	if (GetLastError() != ERROR_IO_PENDING)
	{
		VmxPatchDebugPrint(
			"KdpWritePipe(): WARNING: Write failed (%lu)\n",
			GetLastError());

		KdpDisconnectPipe();

		return FALSE;
	}

	if (WaitForSingleObject(
		g_VmxKdPipeEvent,
		WRITE_TIMEOUT) != WAIT_OBJECT_0)
	{
		VmxPatchDebugPrint(
			"KdpWritePipe(): WARNING: Write timed out, canceling I/O\n");

		CancelIo(
			g_VmxKdPipe
			);
	}

	if (!GetOverlappedResult(
		g_VmxKdPipe,
		&Overlapped,
		&Written,
		TRUE))
	{
		VmxPatchDebugPrint(
			"KdpWritePipe(): WARNING: Write failed (pending) (%lu)\n",
			GetLastError());

		KdpDisconnectPipe();

		return FALSE;
	}

	if (Written < Length)
	{
		VmxPatchDebugPrint(
			"KdpWritePipe(): WARNING: Short write %lu (%lu expected)\n",
			Written,
			Length
			);

		return FALSE;
	}

	return TRUE;
}

KD_RECV_CODE
KdpReadPipe(
	__out PUCHAR Buf,
	__in ULONG Length,
	__in BOOLEAN AllowTimeout
	)
{
	OVERLAPPED Overlapped;
	ULONG      Read;
	ULONG      Timeout;
	BOOLEAN    TimedOut;
	PUCHAR     OrigBuf;
	ULONG      OrigLength;

	TimedOut   = FALSE;
	OrigBuf    = Buf;
	OrigLength = Length;

	if (AllowTimeout)
		Timeout = READ_TIMEOUT;
	else
		Timeout = 0;

	while (Length)
	{
		if (!g_PipeConnected &&
			!KdpConnectPipe(Timeout))
		{
			VmxPatchDebugPrint(
				"KdpReadPipe(): Pipe is not connected\n"
				);

			return KD_RECV_CODE_FAILED;
		}

		ZeroMemory(
			&Overlapped,
			sizeof( OVERLAPPED )
			);

		Overlapped.hEvent = g_VmxKdPipeEvent;

		ResetEvent(
			g_VmxKdPipeEvent
			);

		Read = 0;

		if (ReadFile(
			g_VmxKdPipe,
			Buf,
			Length,
			&Read,
			&Overlapped))
		{
			if (Read < Length)
			{
				VmxPatchDebugPrint(
					"KdpReadPipe(): Short read %lu (%lu expected)\n",
					Read,
					Length
					);

				Buf    += Read;
				Length -= Read;

				continue;
			}

			if (g_LogPackets)
			{
				VmxPatchDebugPrint(
					"KdpReadPipe(): Read %lu bytes from KD:\n",
					Length
					);

				dumphex((char*)OrigBuf, OrigLength, 4);
			}

			return KD_RECV_CODE_OK;
		}

		if (GetLastError() != ERROR_IO_PENDING)
		{
			VmxPatchDebugPrint(
				"KdpReadPipe(): WARNING: Read failed (%lu)\n",
				GetLastError());

			KdpDisconnectPipe();

			return KD_RECV_CODE_FAILED;
		}

		if (WaitForSingleObject(
			g_VmxKdPipeEvent,
			Timeout) != WAIT_OBJECT_0)
		{
			if (AllowTimeout)
			{
				VmxPatchDebugPrint(
					"KdpReadPipe(): WARNING: Read timed out, canceling I/O\n");
			}

			CancelIo(
				g_VmxKdPipe
				);

			TimedOut = TRUE;
		}

		if (!GetOverlappedResult(
			g_VmxKdPipe,
			&Overlapped,
			&Read,
			TRUE))
		{
			if (AllowTimeout || !TimedOut)
			{
				VmxPatchDebugPrint(
					"KdpReadPipe(): WARNING: Read failed (pending) (%lu)\n",
					GetLastError());
			}

			if (TimedOut)
				return KD_RECV_CODE_TIMEOUT;
			else
				KdpDisconnectPipe();

			return KD_RECV_CODE_FAILED;
		}

		if (Read < Length)
		{
			VmxPatchDebugPrint(
				"KdpReadPipe(): Short read %lu (%lu expected)\n",
				Read,
				Length
				);

			Buf    += Read;
			Length -= Read;

			continue;
		}

		if (g_LogPackets)
		{
			VmxPatchDebugPrint(
				"KdpReadPipe(): Read %lu bytes from KD:\n",
				Length
				);

			dumphex((char*)OrigBuf, OrigLength, 4);
		}

		return KD_RECV_CODE_OK;
	}

	return KD_RECV_CODE_OK;
}

VOID
KdCompPutByte(
	__in UCHAR Byte
	)
{
	(VOID) KdpWritePipe(
		&Byte,
		1
		);
}

KD_RECV_CODE
KdCompGetByte(
	__out PUCHAR Byte
	)
{
	return KdpReadPipe(
		Byte,
		1,
		TRUE
		);
}

KD_RECV_CODE
KdCompPollByte(
	__out PUCHAR Byte
	)
{
	return KdpReadPipe(
		Byte,
		1,
		FALSE
		);
}

VOID
KdpSendString(
	__in CONST UCHAR *Data,
	__in ULONG Length
	)
{
	/*
	while (Length > 0)
	{
		KdCompPutByte(
			Data
			);

		Data   += 1;
		Length -= 1;
	}
	*/

	(VOID) KdpWritePipe(
		Data,
		Length
		);
}

KD_RECV_CODE
KdpReceiveString(
	__out PUCHAR Data,
	__in ULONG Length
	)
{
/*	UCHAR        Char;
	KD_RECV_CODE RecvCode;

	while (Length > 0)
	{
		RecvCode = KdCompGetByte(
			&Char
			);

		if (RecvCode != KD_RECV_CODE_OK)
			return RecvCode;

		*Data++ = Char;
		Length -= 1;
	}

	return KD_RECV_CODE_OK;*/

	return KdpReadPipe(
		Data,
		Length,
		TRUE
		);
}

VOID
KdpSendControlPacket(
	__in ULONG PacketType,
	__in_opt ULONG AckPacketId
	)
{
	KD_PACKET Packet;

	Packet.PacketLeader = CONTROL_PACKET_LEADER;

	//
	// KDCOM BUG: We send a garbage PacketId if we don't have an explicit ack
	// PacketId.
	//

	if (AckPacketId)
		Packet.PacketId = AckPacketId;

	Packet.PacketType = (USHORT)PacketType;
	Packet.ByteCount  = 0;
	Packet.Checksum   = 0;

	KdpSendString(
		(CONST UCHAR*)&Packet,
		sizeof( KD_PACKET )
		);
}

ULONG
KdpComputeChecksum(
	IN CONST VOID* Buffer,
	IN ULONG Length
	)
{
	CONST UCHAR* CheckSumData = (CONST UCHAR*)Buffer;
	ULONG CheckSum = 0;

	while(Length--)
		CheckSum += *CheckSumData++;

	return CheckSum;
}

VOID
KdSendPacket(
	__in ULONG PacketType,
	__in PKD_BUFFER PacketData,
	__in_opt PKD_BUFFER SecondaryData,
	__inout PKD_CONTEXT KdContext
	)
{
	KD_PACKET Packet;

	//
	// If we've got secondary data, first compute the checksum over it.
	// Otherwise, initialize the checksum to zero.
	//

	if (SecondaryData)
	{
		Packet.Checksum = KdpComputeChecksum(
			SecondaryData->Data,
			SecondaryData->Length
			);
	}
	else
	{
		Packet.Checksum = 0;
	}

	//
	// Now compute the checksum for the primary data and add it in.
	//

	Packet.Checksum += KdpComputeChecksum(
		PacketData->Data,
		PacketData->Length
		);

	//
	// Reset our retry count.
	//

	KdCompNumberRetries = KdCompRetryCount;

	//
	// Initialize the packet.
	//

	Packet.ByteCount = PacketData->Length;

	if (SecondaryData)
		Packet.ByteCount += SecondaryData->Length;

	Packet.PacketLeader = PACKET_LEADER;
	Packet.PacketType   = (USHORT)PacketType;

	for (;;)
	{
		if (g_ShutDown)
			return;

		//
		// If we are out of retries, check for certain special apis that are
		// allowed to be dropped if we don't get a reply.  These cause the
		// debugger to become disconnected.
		// 

		if (!KdCompNumberRetries)
		{
			VmxPatchDebugPrint(
				"KdSendPacket(%x): Reached zero retries, check for droppable packet exemption\n",
				PacketType
				);

			switch (PacketType)
			{

			case PACKET_TYPE_KD_DEBUG_IO:
				//
				// If we're a PrintString request, unset debugger present if
				// we are retrying.  This is because things use a zero length
				// PrintString request to detect if the debugger is present.
				//
				if (*(PULONG)PacketData->Data == DbgKdPrintStringApi)
				{
					*KdContext->Flags &= ~KDVM_FLAG_CONNECT_DEBUGGER;
					*KdContext->Flags |=  KDVM_FLAG_DISCONNECT_DEBUGGER;

					VmxPatchDebugPrint(
						"KdSendPacket(%x): Disconnect debugger and abort for DbgKdPrintStringApi\n",
						PacketType
						);

//					*KdDebuggerNotPresent            = TRUE;
//					SharedUserData->DebuggerEnabled &= ~2;
					KdCompNextPacketIdToSend         = INITIAL_PACKET_ID | SYNC_PACKET_ID;
					KdCompPacketIdExpected           = INITIAL_PACKET_ID;
					return;
				}
				break;

			case PACKET_TYPE_KD_STATE_CHANGE:
			case PACKET_TYPE_KD_STATE_CHANGE64:
				if (*(PULONG)PacketData->Data == DbgKdLoadSymbolsStateChange)
				{
					*KdContext->Flags &= ~KDVM_FLAG_CONNECT_DEBUGGER;
					*KdContext->Flags |=  KDVM_FLAG_DISCONNECT_DEBUGGER;

					VmxPatchDebugPrint(
						"KdSendPacket(%x): Disconnect debugger and abort for DbgKdLoadSymbolsStateChange\n",
						PacketType
						);

//					*KdDebuggerNotPresent            = TRUE;
//					SharedUserData->DebuggerEnabled &= ~2;
					KdCompNextPacketIdToSend         = INITIAL_PACKET_ID | SYNC_PACKET_ID;
					KdCompPacketIdExpected           = INITIAL_PACKET_ID;
					return;
				}
				break;

			case PACKET_TYPE_KD_FILE_IO:
				if (*(PULONG)PacketData->Data == 0x3430) // XXX: Need name!
				{
					*KdContext->Flags &= ~KDVM_FLAG_CONNECT_DEBUGGER;
					*KdContext->Flags |=  KDVM_FLAG_DISCONNECT_DEBUGGER;

					VmxPatchDebugPrint(
						"KdSendPacket(%x): Disconnect debugger and abort for 0x3430 (FILE_IO)\n",
						PacketType
						);

//					*KdDebuggerNotPresent            = TRUE;
//					SharedUserData->DebuggerEnabled &= ~2;
					KdCompNextPacketIdToSend         = INITIAL_PACKET_ID | SYNC_PACKET_ID;
					KdCompPacketIdExpected           = INITIAL_PACKET_ID;
					return;
				}
				break;

			default:
				break;

			}
		}

		VmxPatchDebugPrint(
			"KdSendPacket(%x): Try send packet with packet ID %08x\n",
			PacketType,
			KdCompNextPacketIdToSend
			);

		//
		// Set the packet id up and send the header and any data that needs to
		// be transmitted.
		//

		Packet.PacketId = KdCompNextPacketIdToSend;

		KdpSendString(
			(CONST UCHAR*)&Packet,
			sizeof( KD_PACKET )
			);
		KdpSendString(
			PacketData->Data,
			PacketData->Length
			);

		if (SecondaryData)
		{
			KdpSendString(
				SecondaryData->Data,
				SecondaryData->Length
				);
		}

		//
		// Transmit the trailing byte.
		//

		KdCompPutByte(
			PACKET_TRAILING_BYTE
			);

		//
		// Receive an ACK packet.
		//

		switch (KdReceivePacket(
			PACKET_TYPE_KD_ACKNOWLEDGE,
			0,
			0,
			0,
			KdContext))
		{

		case KD_RECV_CODE_TIMEOUT:
			KdCompNumberRetries -= 1;
			//
			// Fall through.
			//
		case KD_RECV_CODE_FAILED:

			VmxPatchDebugPrint(
				"KdSendPacket(%x): Send FAILED or TIMED OUT, retry %lu\n",
				PacketType,
				KdCompNumberRetries
				);

			continue;

		default:
		case KD_RECV_CODE_OK:

			//
			// Okay, we're good.  Let's consider ourselves successful and
			// move on.
			//

			KdCompNextPacketIdToSend &= ~SYNC_PACKET_ID;

			KdCompRetryCount = KdContext->RetryCount;

			VmxPatchDebugPrint(
				"KdSendPacket(%x): Sent packet successfully, NextPacketIdToSend now %08x\n",
				PacketType,
				KdCompNextPacketIdToSend
				);

			return;
		}
	}
}

KD_RECV_CODE
KdCompReceivePacketLeader(
	__in ULONG PacketType,
	__out PULONG PacketLeader,
	__inout PKD_CONTEXT KdContext
	)
{
	UCHAR   Char;
	UCHAR   LastChar;
	ULONG   Index;
	BOOLEAN BreakInRequested;

	LastChar          = 0;
	Index             = 0;
	BreakInRequested  = FALSE;

	for (;;)
	{
		if (g_ShutDown)
			return KD_RECV_CODE_FAILED;

		//
		// Receive a character.
		//

		switch (KdCompGetByte(
			&Char))
		{

		default:
		case KD_RECV_CODE_OK:
			break;

		case KD_RECV_CODE_TIMEOUT:
			if (!BreakInRequested)
			{
				VmxPatchDebugPrint(
					"KdCompReceivePacketLeader(): Timeout\n");

				return KD_RECV_CODE_TIMEOUT;
			}
			else
			{
				KdContext->BreakInRequested = TRUE;

				VmxPatchDebugPrint(
					"KdCompReceivePacketLeader(): return BREAK IN REQUESTED (timeout)\n"
					);

				return KD_RECV_CODE_FAILED;
			}
			break;

		case KD_RECV_CODE_FAILED:
			Index = 0;
			continue;

		}

		//
		// Okay, we've got a character.  Figure out what to be doing with it
		// now.
		//

		switch (Char)
		{


		case PACKET_LEADER_BYTE:
		case CONTROL_PACKET_LEADER_BYTE:
			//
			// If it's the first character in the sequence, then we'll save the
			// expected value, set the index to one, and receive the next.
			//

			if (Index == 0)
			{
				LastChar = Char;
				Index    = 1;

				continue;
			}
			else
			{
				//
				// Otherwise, determine if we got a different sequence.  If so,
				// then switch us over and set the index to one, otherwise
				// increment the index.
				//

				if (LastChar != Char)
					Index  = 1;
				else
					Index += 1;

				//
				// If we don't have a complete leader yet, then receive the
				// next character.
				//

				if (Index < 4)
					continue;

				//
				// Okay, we've got the leader.  We're done.  If we saw a
				// breakin packet just before we got the valid leader, then set
				// the breakin requested flag.  Also, update the debugger state
				// to reflect that the debugger is now connected.
				//

				if (BreakInRequested)
					KdContext->BreakInRequested = TRUE;

				if (Char == CONTROL_PACKET_LEADER_BYTE)
					*PacketLeader = CONTROL_PACKET_LEADER;
				else
					*PacketLeader = PACKET_LEADER_BYTE;

				*KdContext->Flags &= ~KDVM_FLAG_DISCONNECT_DEBUGGER;
				*KdContext->Flags |=  KDVM_FLAG_CONNECT_DEBUGGER;

//				*KdDebuggerNotPresent              = FALSE;
//				SharedUserData->KdDebuggerEnabled |= 2;

				return KD_RECV_CODE_OK;
			}
			break;

		default:
			//
			// N.B. This means that only if the last byte was a break in, we'll
			// consider it really a break in request.
			//

			BreakInRequested = (Char == BREAKIN_PACKET_BYTE);
			Index            = 0;
			continue;
		}
	}

	UNREFERENCED_PARAMETER( PacketType );
}

KD_RECV_CODE
KdReceivePacket(
	__in ULONG PacketType,
	__inout_opt PKD_BUFFER PacketData,
	__inout_opt PKD_BUFFER SecondaryData,
	__out_opt PULONG PayloadBytes,
	__inout_opt PKD_CONTEXT KdContext
	)
{
	UCHAR        Char;
	KD_PACKET    Packet;
	KD_RECV_CODE RecvCode;
	ULONG        Checksum;

	//
	// If we are the special request for a break in poll, then handle that now.
	//

	if (PacketType == PACKET_TYPE_KD_BREAKIN_POLL)
	{
		if ((KdCompPollByte(
			 &Char) != KD_RECV_CODE_OK)    ||
			(Char != BREAKIN_PACKET_BYTE))
		{
			VmxPatchDebugPrint("KdReceivePacket(PACKET_TYPE_KD_BREAKIN_POLL): Returning KD_RECV_CODE_TIMEOUT.\n");
			return KD_RECV_CODE_TIMEOUT;
		}

		VmxPatchDebugPrint("KdReceivePacket(PACKET_TYPE_KD_BREAKIN_POLL): Got breakin request, return OK\n");

		return KD_RECV_CODE_OK;
	}

	//
	// Otherwise, we've got work to do.  Let's get to it...
	//

	for (;;)
	{
		//
		// Receive a packet leader.
		//

		if (g_ShutDown)
			return KD_RECV_CODE_FAILED;

		RecvCode = KdCompReceivePacketLeader(
			PacketType,
			&Packet.PacketLeader,
			KdContext
			);

		//
		// If we didn't get a timeout status, then reset the retry count.
		//

		if (RecvCode != KD_RECV_CODE_TIMEOUT)
		{
			KdCompNumberRetries = KdCompRetryCount;

			VmxPatchDebugPrint("KdReceivePacket(%x): Reset retry count\n",
				PacketType
				);
		}

		//
		// If we didn't successfully receive the leader, then bail out now.
		//

		if (RecvCode != KD_RECV_CODE_OK)
		{
			VmxPatchDebugPrint("KdReceivePacket(%x): Receive leader failed (%x), return.\n",
				PacketType,
				RecvCode
				);

			return RecvCode;
		}

		for (;;)
		{
			//
			// Now we'll need to receive the packet type and length.
			//

			RecvCode = KdpReceiveString(
				(PUCHAR)&Packet.PacketType,
				2
				);

			if (RecvCode == KD_RECV_CODE_TIMEOUT)
				return KD_RECV_CODE_TIMEOUT;
			else if (RecvCode == KD_RECV_CODE_FAILED)
				break;

			VmxPatchDebugPrint("KdReceivePacket(%x): Got packet type %x\n",
				PacketType,
				Packet.PacketType
				);

			//
			// If we're a control packet requesting a resend, then return a
			// failure status right now.  If we're being called by
			// KdSendPacket, then it will handle this by resending the packet.
			// If we're not in KdSendPacket, then something is not right if we
			// got a resend request.
			//

			if (Packet.PacketLeader == CONTROL_PACKET_LEADER &&
				Packet.PacketType   == PACKET_TYPE_KD_RESEND)
			{
				VmxPatchDebugPrint("KdReceivePacket(%x): Fail request due to control resend request\n",
					PacketType
					);

				return KD_RECV_CODE_FAILED;
			}

			//
			// Not a resend request, keep on receiving the packet header.
			//

			RecvCode = KdpReceiveString(
				(PUCHAR)&Packet.ByteCount,
				2
				);

			if (RecvCode == KD_RECV_CODE_TIMEOUT)
				return KD_RECV_CODE_TIMEOUT;
			else if (RecvCode == KD_RECV_CODE_FAILED)
				break;

			VmxPatchDebugPrint("KdReceivePacket(%x): Got packet length %lu\n",
				PacketType,
				Packet.ByteCount
				);

			RecvCode = KdpReceiveString(
				(PUCHAR)&Packet.PacketId,
				4
				);

			if (RecvCode == KD_RECV_CODE_TIMEOUT)
				return KD_RECV_CODE_TIMEOUT;
			else if (RecvCode == KD_RECV_CODE_FAILED)
				break;

			VmxPatchDebugPrint("KdReceivePacket(%x): Got packet ID %08x\n",
				PacketType,
				Packet.PacketId
				);

			RecvCode = KdpReceiveString(
				(PUCHAR)&Packet.Checksum,
				4
				);

			if (RecvCode == KD_RECV_CODE_TIMEOUT)
				return KD_RECV_CODE_TIMEOUT;
			else if (RecvCode == KD_RECV_CODE_FAILED)
				break;


			VmxPatchDebugPrint("KdReceivePacket(%x): Got packet checksum %08x\n",
				PacketType,
				Packet.Checksum
				);

			break;
		}

		//
		// If we failed to receive the packet, then determine the course of
		// action.  For a normal packet, this may involve requesting a resend
		// before retrying the receive operation.
		//

		if (RecvCode == KD_RECV_CODE_FAILED)
		{
			VmxPatchDebugPrint("KdReceivePacket(%x): Receiving header failed, leader %08X\n",
				PacketType,
				Packet.PacketLeader
				);

			//
			// If we were receiving a control packet, then just try receiving
			// again.
			//

			if (Packet.PacketLeader == CONTROL_PACKET_LEADER)
				continue;

			//
			// Otherwise, request a resend.
			//

			KdpSendControlPacket(
				PACKET_TYPE_KD_RESEND,
				0
				);
			continue;
		}

		//
		// Okay, we've got a complete packet header.  Let's determine what to
		// do with it now.  The first thing to check is whether we're a control
		// or normal packet.
		//

		if (Packet.PacketLeader != CONTROL_PACKET_LEADER)
		{
			//
			// If we got a normal packet of type acknowledge, then this is a
			// special case.  Handle it now.
			//

			VmxPatchDebugPrint(
				"KdReceivePacket(%x): Got NORMAL PACKET %x\n",
				PacketType,
				Packet.PacketType
				);

			if (PacketType == PACKET_TYPE_KD_ACKNOWLEDGE)
			{
				//
				// Check that the packet id matches the last packet id.  If not
				// then we need to ask for a resend, but if we do need to ask
				// for a resend, we'll adjust our expected packet id and return
				// a success status, not waiting for a reply.
				//

				if (Packet.PacketId == KdCompPacketIdExpected)
				{
					KdpSendControlPacket(
						PACKET_TYPE_KD_RESEND,
						0
						);

					KdCompNextPacketIdToSend ^= 1;

					VmxPatchDebugPrint(
						"KdReceivePacket(%x): Got NORMAL ACKNOWLEDGE for expected packet ID %08X, NextPacketIdToSend is now %08X, RESEND sent, return OK\n",
						PacketType,
						Packet.PacketId,
						KdCompNextPacketIdToSend
						);

					return KD_RECV_CODE_OK;
				}
				else
				{
					//
					// Otherwise acknowledge their packet and continue with our
					// receiving.  Note that we send a control acknowledge and
					// not a "normal" acknowledge.
					//

					KdpSendControlPacket(
						PACKET_TYPE_KD_ACKNOWLEDGE,
						Packet.PacketId
						);

					VmxPatchDebugPrint(
						"KdReceivePacket(%x): Got NORMAL ACKNOWLEDGE for other packet ID %08X (Expected %08X), send ACKNOWLEDGE and continue\n",
						PacketType,
						Packet.PacketId,
						KdCompPacketIdExpected
						);

					continue;
				}
			}

			//
			// If the length is greater than the maximum byte count, then we
			// know that something's bad.  Request a resend.  Similarly, if we
			// get a packet that is less than the expected packet size, then
			// we'll also request a resend.  Also, if the packet size is above
			// the maximum optional data size that the caller has allowed for
			// it, then that is also illegal and grounds for a resend request.
			//

			if ((Packet.ByteCount > PACKET_MAX_SIZE                                         ) ||
			    (Packet.ByteCount < PacketData->MaximumLength                               ) ||
				(Packet.ByteCount > SecondaryData->MaximumLength + PacketData->MaximumLength))
			{
				KdpSendControlPacket(
					PACKET_TYPE_KD_RESEND,
					0
					);

				VmxPatchDebugPrint(
					"KdReceivePacket(%x): Packet length is wrong (%lu, %lu, %lu), send RESEND and continue\n",
					PacketType,
					Packet.ByteCount,
					PacketData->MaximumLength,
					SecondaryData->MaximumLength
					);

				continue;
			}

			//
			// Okay, the length checks out.  Let's receive data.  We may have
			// some variable length payload data at the end; if so, we'll need
			// to receive that as well.
			//

			*PayloadBytes = Packet.ByteCount - PacketData->MaximumLength;

			if (KdpReceiveString(
				PacketData->Data,
				PacketData->MaximumLength) != KD_RECV_CODE_OK)
			{
				KdpSendControlPacket(
					PACKET_TYPE_KD_RESEND,
					0
					);

				VmxPatchDebugPrint(
					"KdReceivePacket(%x): Failed to receive primary packet data, send RESEND and continue\n",
					PacketType
					);

				continue;
			}

			PacketData->Length = PacketData->MaximumLength;

			if (KdpReceiveString(
				SecondaryData->Data,
				*PayloadBytes) != KD_RECV_CODE_OK)
			{
				KdpSendControlPacket(
					PACKET_TYPE_KD_RESEND,
					0
					);

				VmxPatchDebugPrint(
					"KdReceivePacket(%x): Failed to receive secondary packet data, send RESEND and continue\n",
					PacketType
					);

				continue;
			}

			SecondaryData->Length = (USHORT)*PayloadBytes;

			//
			// We've got all the data, so let's receive the trailing byte to
			// ensure that we're still all in synch here.
			//

			if ((KdCompGetByte(
				&Char) != KD_RECV_CODE_OK) ||
				(Char != PACKET_TRAILING_BYTE))
			{
				KdpSendControlPacket(
					PACKET_TYPE_KD_RESEND,
					0
					);

				VmxPatchDebugPrint(
					"KdReceivePacket(%x): Failed to receive trailing byte, send RESEND and continue\n",
					PacketType
					);

				continue;
			}

			//
			// At this point, we have successfully received the entire (normal)
			// packet.  Let's ensure that the packet type matches the expected
			// packet type, and that our packet IDs match up.
			//

			if (Packet.PacketType != PacketType)
			{
				//
				// Hrm.  We've got a packet that appears to be valid, but it
				// wasn't of the type that we were expecting.  Acknowledge it
				// and receive some more.
				//

				KdpSendControlPacket(
					PACKET_TYPE_KD_ACKNOWLEDGE,
					Packet.PacketId
					);

				VmxPatchDebugPrint(
					"KdReceivePacket(%x): Got unexpected packet type %x (packet ID %08X), send ACKNOWLEDGE and continue\n",
					PacketType,
					Packet.PacketType,
					Packet.PacketId
					);

				continue;
			}

			//
			// Make sure that the packet id is legal.  If not, then we'll
			// request a resend.
			//

			if ((Packet.PacketId != INITIAL_PACKET_ID    ) &&
				Packet.PacketId != (INITIAL_PACKET_ID ^ 1))
			{
				//
				// Not a legal packet id.  Request a resend.
				//

				KdpSendControlPacket(
					PACKET_TYPE_KD_RESEND,
					0
					);

				VmxPatchDebugPrint(
					"KdReceivePacket(%x): Got illegal packet ID %08X, send RESEND and continue\n",
					PacketType,
					Packet.PacketId
					);

				continue;
			}

			//
			// It's a legal packet id, but we still need to make sure that it
			// is really the packet id that we were expecting.
			//

			if (Packet.PacketId != KdCompPacketIdExpected)
			{
				//
				// Hrm.  It's a legal packet id, but not what we were
				// expecting.  Acknowledge it and continue receiving.
				//

				KdpSendControlPacket(
					PACKET_TYPE_KD_ACKNOWLEDGE,
					Packet.PacketId
					);

				VmxPatchDebugPrint(
					"KdReceivePacket(%x): Got unexpected packet ID %08X, send ACKNOWLEDGE and continue\n",
					PacketType,
					Packet.PacketId
					);

				continue;
			}

			//
			// Okay, let's verify that the checksum matches up.  We've gotta be
			// bulletproof and all, y'know...
			//

			Checksum  = KdpComputeChecksum(
				PacketData->Data,
				PacketData->Length
				);
			Checksum += KdpComputeChecksum(
				SecondaryData->Data,
				SecondaryData->Length
				);

			if (Checksum != Packet.Checksum)
			{
				//
				// Checksum is bad.  We'll request a resend.
				//

				KdpSendControlPacket(
					PACKET_TYPE_KD_RESEND,
					0
					);

				VmxPatchDebugPrint(
					"KdReceivePacket(%x): Got invalid checksum %08X (should be %08X), send RESEND and continue\n",
					PacketType,
					Packet.Checksum,
					Checksum
					);

				continue;
			}

			//
			// Okay, we'll acknowledge the packet and adjust our expected next
			// packet id accordingly.  This is the end of our long trek and we
			// are victorious!
			//

			KdpSendControlPacket(
				PACKET_TYPE_KD_ACKNOWLEDGE,
				Packet.PacketId
				);

			KdCompPacketIdExpected ^= 1;

			VmxPatchDebugPrint(
				"KdReceivePacket(%x): Successfully received packet (expected packet ID now %08X), send ACKNOWLEDGE and return OK\n",
				PacketType,
				KdCompPacketIdExpected
				);

			return KD_RECV_CODE_OK;
		}
		else
		{
			//
			// We've got a control packet.  Let's deal with it.
			//

			VmxPatchDebugPrint(
				"KdReceivePacket(%x): Got CONTROL PACKET %x\n",
				PacketType,
				Packet.PacketType
				);

			switch (Packet.PacketType)
			{

			case PACKET_TYPE_KD_ACKNOWLEDGE:
				//
				// If it's an acknowledge packet, make sure that it corresponds
				// to the proper packet id.  Otherwise, that's bad and we
				// should ignore it.
				//

				if ((KdCompNextPacketIdToSend & ~SYNC_PACKET_ID) != Packet.PacketId)
				{
					VmxPatchDebugPrint(
						"KdReceivePacket(%x): Got bad packet id %08X (NextPacketIdToSend %08X) for ACKNOWLEDGE, continue\n",
						PacketType,
						Packet.PacketId,
						KdCompNextPacketIdToSend & ~SYNC_PACKET_ID
						);

					continue;
				}

				//
				// If we got an acknowledge packet but we weren't looking for
				// one, then we should ignore it.
				//

				if (PacketType != PACKET_TYPE_KD_ACKNOWLEDGE)
				{
					VmxPatchDebugPrint(
						"KdReceivePacket(%x): Was expecting packet %x but got ACKNOWLEDGE, continue\n",
						PacketType,
						PacketType
						);

					continue;
				}

				//
				// Otherwise, we were wanting an acknowledgement and we got it.
				// That means we're done (but we need to update our packet id).
				// Hooray!
				//

				KdCompNextPacketIdToSend ^= 1;

				VmxPatchDebugPrint(
					"KdReceivePacket(%x): Got CONTROL ACKNOWLEDGE, NextPacketIdToSend now %08x, return OK\n",
					PacketType,
					KdCompNextPacketIdToSend
					);

				return KD_RECV_CODE_OK;

			case PACKET_TYPE_KD_RESEND:
				//
				// Note that this appears to be a redundant check as we already
				// handled the case of CONTROL RESEND right after we received
				// the packet type.  We have the same handling here, which is
				// to return a failure status.  KdSendPacket will interpret
				// this as a request to resend.
				//

				VmxPatchDebugPrint(
					"KdReceivePacket(%x): Got CONTROL RESEND, return FAILED\n",
					PacketType
					);

				return KD_RECV_CODE_FAILED;

			case PACKET_TYPE_KD_RESET:
				//
				// A reset packet was received.  This means that we should
				// change our states to known values and acknowledge the reset
				// by replying to the requestor with another reset packet.
				//

				KdCompNextPacketIdToSend   = INITIAL_PACKET_ID;
				KdCompPacketIdExpected     = INITIAL_PACKET_ID;

				KdpSendControlPacket(
					PACKET_TYPE_KD_RESET,
					0
					);

				//
				// A reset packet counts as a universal failure.
				//

				VmxPatchDebugPrint(
					"KdReceivePacket(%x): Got RESET, return FAILED\n",
					PacketType
					);

				return KD_RECV_CODE_FAILED;

			default:
				//
				// We received an unrecognized packet type as a control packet.
				// This is bad.  We should ignore it and keep on receiving.
				//

				VmxPatchDebugPrint(
					"KdReceivePacket(%x): Got other CONTROL PACKET %x, continue\n",
					PacketType,
					Packet.PacketType
					);

				continue;

			}
		}
	}


}
