#ifdef _MSC_VER
#pragma once
#endif

#ifndef _VMXPATCH_H
#define _VMXPATCH_H

typedef enum _KD_RECV_CODE
{
	KD_RECV_CODE_OK       = 0,
	KD_RECV_CODE_TIMEOUT  = 1,
	KD_RECV_CODE_FAILED   = 2
} KD_RECV_CODE, * PKD_RECV_CODE;

#define KDVM_SIGNATURE_INITIALIZING "KdVm Communications Back Channel Interface - (C) 2007 Skywing\x00\xAA\00"
#define KDVM_SIGNATURE_ESTABLISHED  "KdVm Communications Back Channel Interface - (C) 2007 Skywing\x00\x55\00"
#define KDVM_SIGNATURE_LENGTH 64

#define KD_MAX_DATA_BLOCK_SIZE 4000

#define KDVM_FLAG_BREAK_IN_REQUESTED   0x00000001
#define KDVM_FLAG_DISCONNECT_DEBUGGER  0x00000002
#define KDVM_FLAG_CONNECT_DEBUGGER     0x00000004

#define BUGCODE_KDVMWARE               0x0401BABE
#define BUGID_KDVMWARE_LARGE_RECV      0x00000000
#define BUGID_KDVMWARE_LARGE_SEND      0x00000001

typedef struct _KDVM_BUFFER_DESCRIPTOR
{
	ULONG        PacketType;
	USHORT       PrimaryLength;
	USHORT       SecondaryLength;
	ULONG        Flags;
	KD_RECV_CODE RecvCode;
	UCHAR        Buffer[ KD_MAX_DATA_BLOCK_SIZE ];
} KDVM_BUFFER_DESCRIPTOR, * PKDVM_BUFFER_DESCRIPTOR;

typedef enum _KDVM_OPERATION
{
	KdVmOperationNone,
	KdVmOperationInitialize,
	KdVmOperationReceivePacket,
	KdVmOperationSendPacket,
	KdVmOperationBugCheck,

	LastKdVmOperation
} KDVM_OPERATION, * PKDVM_OPERATION;

typedef struct _KDVM_DATA_BLOCK
{
	//
	// Magic prefix for searching for virtual VM physical page.
	//

	UCHAR                  Magic[ KDVM_SIGNATURE_LENGTH ];

	//
	// Structure length.
	//

	ULONG                  Length;

	//
	// Why we're requesting attention from the host.
	//

	KDVM_OPERATION         Operation;

	//
	// Host connection activity indicator.
	//

	BOOLEAN                HostLinkActive;

	//
	// Bug check request from host to guest code in case we get into a bad
	// state and need to get a dump to figure out what's going on.
	//

	BOOLEAN                BugCheckRequested;

	UCHAR                  Spare[ 2 ];

	//
	// Operational parameters.
	//

	KDVM_BUFFER_DESCRIPTOR Parameters;
} KDVM_DATA_BLOCK, * PKDVM_DATA_BLOCK;

#ifndef PAGE_SIZE
#define PAGE_SIZE 4096
#endif

C_ASSERT( sizeof(KDVM_DATA_BLOCK) <= PAGE_SIZE );

//
// support routines
//

void
VmxPatchDebugPrint(
	__in __format_string const char *Fmt,
	...
	);

void dumphex(const char *buf, int len, int pos);

extern HANDLE g_VmxKdPipe;
extern HANDLE g_VmxKdPipeEvent;
extern BOOLEAN g_PipeConnected;
extern volatile BOOLEAN g_LogPackets;
extern volatile BOOLEAN g_ShutDown;

//
// kdcom support
//

typedef struct _KD_BUFFER
{
	USHORT  Length;
	USHORT  MaximumLength;
	PUCHAR  Data;
} KD_BUFFER, * PKD_BUFFER;

typedef struct _KD_CONTEXT
{
	ULONG      RetryCount;
	BOOLEAN    BreakInRequested;

	//
	// Extensions
	//

	HANDLE     Pipe;
	OVERLAPPED Overlapped;
	PULONG     Flags;
} KD_CONTEXT, * PKD_CONTEXT;

VOID
KdSendPacket(
	__in ULONG PacketType,
	__in PKD_BUFFER PacketData,
	__in_opt PKD_BUFFER SecondaryData,
	__inout PKD_CONTEXT KdContext
	);

KD_RECV_CODE
KdReceivePacket(
	__in ULONG PacketType,
	__inout_opt PKD_BUFFER PacketData,
	__inout_opt PKD_BUFFER SecondaryData,
	__out_opt PULONG PayloadBytes,
	__inout_opt PKD_CONTEXT KdContext
	);

VOID
KdpResetState(
	VOID
	);

#endif
